#include<dune/grid/common/genericreferenceelements.hh>
#include<dune/grid/common/quadraturerules.hh>
#include<dune/pdelab/common/geometrywrapper.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>

/* local operator for solving the Poisson equation
 *
 * -\Delta u = 4 pi \rho  in \Omega 
 *         u = 0          on \partial \Omega
 * with conforming finite elements on all types of grids in any dimension
 * Week formulation in Residual form:
 * int ( grad u * grad v - 4 pi \rho * v ) = 0
 */ 

template<class B, class GFVec>
class PoissonLocalOperator : 
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags
{
public:
  
  // pattern assembly flags
  enum { doPatternVolume = true };

  // residual assembly flags
  enum { doAlphaVolume = true };
  //*************************************************************************
  // if set false: Run-time error: Parameter 4 to routine DLASCLS was incorrect:
  // konsistent reproduzierbar in uni-linux-machine
  //**************************************************************************
  //enum { doAlphaBoundary = false };
  

  // Constructor
  PoissonLocalOperator (const B& b_, const unsigned int Poissintorder_=6)
    : b(b_), Poissintorder(Poissintorder_){}


	
//OLD template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
//OLD void jacobian_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, Dune::PDELab::LocalMatrix<R>& mat) const
template<typename EG, typename LFSU, typename X, typename LFSV, typename M>
void jacobian_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, M& mat) const
{
/* TEST partition type */
	if (eg.entity().partitionType() == 4)
	{
		std::cout << "Ghost Entity entered - should not happen" << std::endl;
		for(unsigned i = 0; i < lfsu.size(); ++i)
		{
			for(unsigned j = 0; j < lfsu.size(); ++j)
			{
				mat.accumulate(lfsu,i,lfsv,j,0.0);
 			}
		}
		return ;
	}
//*/	
	/*
	else 
	{	if (eg.entity().partitionType() == 1)
		{
			std::cout << "I am border entity " << std::endl;
		}
		else
		{
			if (eg.entity().partitionType() == 0)
			{
				std::cout << "I am interior entity " << std::endl;
			}
		}
	}			
	std::cout << "My partition Type is " << eg.entity().partitionType() << std::endl;	
	*/
	// Typedefs, LBU for LocalBasis of function space U
  typedef typename LFSU::Traits::FiniteElementType::Traits::LocalBasisType::Traits LBUTraits; 
	typedef typename LBUTraits::DomainFieldType DF; 
	typedef typename LBUTraits::DomainType Domain;
  static const unsigned dimD = LBUTraits::dimDomain;
  typedef typename LBUTraits::RangeFieldType RF;
  typedef typename LBUTraits::RangeType Range;
  static const unsigned dimR = LBUTraits::dimRange;
  typedef typename LBUTraits::JacobianType JacobianType;
	typedef typename LFSU::Traits::SizeType size_type;
  
	// dimensions
  const int dim = EG::Geometry::dimension;
  const int dimw = EG::Geometry::dimensionworld;
		
  // static checks
  dune_static_assert((Dune::is_same<typename EG::Geometry::ctype, DF>::value),
  "Grids ctype and Finite Elements DomainFieldType must match");

	
  // select quadrature rule
  typedef Dune::QuadratureRule<DF,dimD> QR;
  typedef Dune::QuadratureRules<DF,dimD> QRs;
	Dune::GeometryType gt = eg.geometry().type();
	const QR& rule = QRs::rule(gt,Poissintorder);
		
	RF rho_qp=0.0;
	RF pot_elec_qp=0.0;
	RF Vc_qp=0.0;
	double rs = 0.0;
	double ec_qp = 0.0;
	double denom = 0.0;			
	// loop over quadrature points
  for(typename QR::const_iterator it=rule.begin();
	  it!=rule.end(); ++it) {
		
		// evaluate gradient of basis functions on reference element
	  std::vector<JacobianType> js(lfsu.size());
		//lfsu.localFiniteElement().localBasis().evaluateJacobian(it->position(),js); //?? why no error due to local-only deprecated!! changed at 26.07.2011
  	lfsu.finiteElement().localBasis().evaluateJacobian(it->position(),js);

   	// transform gradients from reference element to real element
   	const Dune::FieldMatrix<DF,dimw,dim> jac = eg.geometry().jacobianInverseTransposed(it->position());
   	std::vector<Dune::FieldVector<RF,dim> > gradphi(lfsu.size());
   	for (size_type i=0; i<lfsu.size(); i++) jac.mv(js[i][0],gradphi[i]);

		// Calculate local matrix contribution from actual quadrature point   
		RF factor = it->weight() * eg.geometry().integrationElement(it->position()); 
		for(unsigned i = 0; i < lfsu.size(); ++i){
			for(unsigned j = 0; j < lfsu.size(); ++j){
				//OLD mat(i,j) += factor * ( gradphi[i] * gradphi[j]);
				mat.accumulate(lfsu,i,lfsv,j,(factor * ( gradphi[i] * gradphi[j])));
				//mat.accumulate(lfsu,i,lfsv,j,1);
			}
		 }
	}  	// for quadratur points
}	// end of method 
	


private:
  const B& b; 
  const unsigned int Poissintorder;
};


 // **********************************************************
  // volume integral depending on test and ansatz functions
  // **********************************************************
  
/*
template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
{
    // extract some types
    typedef typename LFSU::Traits::FiniteElementType::
      Traits::LocalBasisType::Traits::DomainFieldType DF;
    typedef typename LFSU::Traits::FiniteElementType::
      Traits::LocalBasisType::Traits::RangeFieldType RF;
    typedef typename LFSU::Traits::FiniteElementType::
      Traits::LocalBasisType::Traits::JacobianType JacobianType;
    typedef typename LFSU::Traits::FiniteElementType::
      Traits::LocalBasisType::Traits::RangeType RangeType;
    typedef typename LFSU::Traits::SizeType size_type;
        
    // dimensions
    const int dim = EG::Geometry::dimension;
    const int dimw = EG::Geometry::dimensionworld;

    // select quadrature rule
    Dune::GeometryType gt = eg.geometry().type();
    const Dune::QuadratureRule<DF,dim>& rule = Dune::QuadratureRules<DF,dim>::rule(gt,intorder);

    // loop over quadrature points
    for (typename Dune::QuadratureRule<DF,dim>::const_iterator 
           it=rule.begin(); it!=rule.end(); ++it)
    {
    
        // evaluate gradient of basis functions on reference element
        std::vector<JacobianType> js(lfsu.size());
        lfsu.finiteElement().localBasis().evaluateJacobian(it->position(),js);

        // transform gradients from reference element to real element
        const Dune::FieldMatrix<DF,dimw,dim> 
        jac = eg.geometry().jacobianInverseTransposed(it->position());
        std::vector<Dune::FieldVector<RF,dim> > gradphi(lfsu.size());
        for (size_type i=0; i<lfsu.size(); i++)
				{
					jac.mv(js[i][0],gradphi[i]);
				}
        
        // compute gradient of u
        Dune::FieldVector<RF,dim> gradu(0.0);
        for (size_type i=0; i<lfsu.size(); i++) 
				{
					gradu.axpy(x[i],gradphi[i]);
				}

        // evaluatie basis functions on reference element
        std::vector<RangeType> phi(lfsu.size());
        lfsu.finiteElement().localBasis().evaluateFunction(it->position(),phi);
        // compute rho at integration point
        std::vector<typename GFVec::ElementType> rhol(lfsu.size());
        lfsu.vread(rho,rhol);

        RF rho_qp=0.0;
				for (size_type i=0; i<lfsu.size(); i++)
				{
					rho_qp += rhol[i]*phi[i];
				}
        // integrate grad u * grad phi_i + a*u*phi_i - f phi_i
        RF factor = it->weight()*eg.geometry().integrationElement(it->position());
        for (size_type i=0; i<lfsu.size(); i++)
				{
		  // Korrekte Formel, auch bevor Atompositionen eingebaut (Jetzt nur noch PoissonMatrix extrahiert):
		   		r[i] += ( gradu*gradphi[i] - 12.5664*rho_qp*phi[i] )*factor;
				}
			
          //r[i] += ( gradu*gradphi[i] + u*phi[i] )*factor;
		  // 4 * pi not needed in this unit system, sign seams correct, see certik -- wrong!!needed! --?
    }
} 
*/
